local mimetypes = require 'mimetypes'
local sfs = require 'sfs'
local socket = require 'socket'
local pmd3 = require 'pmd3'

--

local usage = require 'argparse' (arg[0], 'Render Paged Markdown 3 documents to PDF.')

usage:argument('input', 'Input markdown file.')
usage:argument('output', 'Output pdf file.'):args '?'
usage:option('--head', 'Use a custom HTML header instead of the internal one\n(useful for fine-tuning PDF theming).')
    :argname '<custom.html>'

usage:add_complete()
local args = usage:parse()

args.output = args.output or sfs.formatPath('%D%b.pdf', args.input)

--

local htm, meta = pmd3.render(assert(sfs.read(args.input)), args.input)
htm = '<head>' .. (args.head and assert(sfs.read(args.head)) or pmd3.head) .. htm .. pmd3.foot

-- web server
local s = socket.bind('127.0.0.1', 2829)
s:settimeout(0.01)

local _webs_run = true
local threads = {}
table.insert(threads, coroutine.create(function()
	while _webs_run do
		local c = s:accept()
		local headers = {}
		local method, path

		if c then
			c:settimeout(0.01)
			table.insert(threads, coroutine.create(function()
				while c do
					local lin, err = c:receive '*l'
					if not lin and err ~= 'timeout' then break end

					if lin and not method then
						method, path = lin:match '^([A-Za-z]+) (/[^ ]*) HTTP'
					elseif lin then
						local k, v = lin:match '^([^:]+):%s*([^\r\n]+)'
						if not k then break end
						headers[k] = v
					end
					coroutine.yield()
				end

				if path == '/' then
					c:send('HTTP/1.1 200 OK\nContent-Type: text/html\n\n' .. htm)
				elseif path then
					local pth = sfs.combine(sfs.getDir(args.input), path:sub(2))
					local f = io.open(pth, 'r')
					if f then
						local mime = mimetypes.guess(path) or 'text/css'
						c:send('HTTP/1.1 200 OK\nContent-Type: ' .. mime .. '\n\n' .. f:read '*a')
						f:close()
					else
						c:send('HTTP/1.1 404 Not Found\n\n')
					end
				end
				c:close()
				coroutine.yield()
			end))
		end
		coroutine.yield()
	end
end))

table.insert(threads, coroutine.create(function()
	local webd = require 'marionette'.start { headless = true, extra_args = '-MOZ_LOG=console:5' }

	webd.navigate 'http://127.0.0.1:2829/'

	while true do
		local lin = webd.stderr:read '*l'
		if lin and lin:find 'console log: "Rendered" %d+ "pages."' then
			break
		end
		coroutine.yield()
	end

	local pagenums = webd._command('WebDriver:ExecuteScript', {
		script = [[
		return Array.from(document.querySelectorAll('.pagedjs_page')).map((page) => (
			[page.getAttribute('pnum-style'), page.hasAttribute('pnum-reset')]
		))]]
	}).value

	print 'rendering complete -> generating pdf'
	coroutine.yield() -- ensure smooth run

	webd.pdf(args.output)
	webd.close()

	_webs_run = false
	coroutine.yield()

	--

	require 'pmd3.pdfmeta'.annotate(args.output, meta or {}, pagenums)
end))

while #threads > 0 do
	for i = #threads, 1, -1 do
		local ok, err = coroutine.resume(threads[i])
		if not ok then
			print(debug.traceback(threads[i], err))
		end
		if coroutine.status(threads[i]) == 'dead' then
			table.remove(threads, i)
		end
	end
end
